/*
** draw.c
**
** Part of RPN calculator
** Copyright (c) 2005, 2006, 2007, 2008, 2010 NuukiaWorld
**
**  RPN calculator is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* includes nokix headers */
#include <nokix/all.h>
#include <math.h>

#include "main.h"
#include "stack.h"
#include "draw.h"

/*******************************************************/

#define  STACK_ROWS     4  // Visible stack rows in normal mode
#define  ROW_HEIGHT     9  // Height of rows in pixels
#define  BASE_HEIGHT    7  // Baseline height in pixels
#define  EDIT_LINE      35 // Vertical position of edit line
#define  ICON_LINE      6  // Vertical position of DEG/RAD/GRAD
#define  CURSOR_WIDTH   6  // Width of the cursor

/*******************************************************/

const unsigned char digit_bits[] = { 4, 3, 1 };
const unsigned char digit_max[] = { 0xF, 0x7, 0x1 };
const unsigned char digit_max_char[] = { 'F', '7', '1' };

static const char angle_units[] = "  DEG\0  RAD\0GRAD";
static const char radix_str[] = "h\0o\0b";

/*******************************************************/

/* converts a double to string using exponent format if
   required for full accuracy
*/
static void dtostr_10 (double value) {

	char *p;
	double d;
	int exp = 0;
	int n;

	// Scale exponent out if required
	d = value;
	if (d != 0) {
		// Scale if >= 1E+8
		if( fabs(d) >= 9.99999995E+7 ) {
   		do {
   			d /= 1E8;
   			exp += 8;
   		} while( fabs(d) >= 9.99999995E+7 );
   		while( fabs(d) >= 9.99999995 ) {
   			d /= 10;
   			exp++;
   		}
      } else {
   		// Scale if < 1E-1
   		while( fabs(d) < 9.99999995E-8 ) {
   			d *= 1E8;
   			exp -= 8;
   		}
   		while( fabs(d) < 9.999999949999999E-1 ) {  // Note: this value avoids problem "10E7"
   			d *= 10;
   			exp--;
   		}
      }
	}

	// Print the mantissa (max. 8 digits)
	ftoa (buf, d);

   // Find least significant digit & end of the string
   n = exp;
   p = buf;
   do {
      if( *p >= '0' ) n--; // *p is one of: -.0123456789
   } while( *++p != '\0' );
   // 10^(n+1) = least significant digit position, p = end of string
   
   if( exp < 0 && n >= -8 ) {
		// Use non-scientific format instead if it can show all digits
		ftoa (buf, value);
	} else if (exp != 0) {
		// Print the exponent if any
		*p++ = 'E';
		if (exp < 0) {
			*p++ = '-';
   		exp = -exp;
		}
		ntoa (exp, p);
 	}
}

/* converts a double to string using specified radix
*/
static void dtostr_radix( double value, bool_t show_radix ) {

	unsigned int digit, l;
   int c;
   char *p;

   buf[0] = '#';
   l = (unsigned int)lround(value);    // Convert to 32-bit integer
   if( radix == RADIX_2 ) l &= 0xFFF;  // Limit binary to 12 digits

   p = &buf[1];
   for( c = 36; c >= 0; c -= digit_bits[radix] ) {
      digit = ( l >> c ) & digit_max[radix];
      if( p > &buf[1] || digit != 0 || c == 0 ) {
        	if( digit > 9 ) digit += 'A' - '9' - 1;
        	*p++ = '0' + digit;
      }
   }
   // Add radix letter to the end if requested
   if(show_radix) *p++ = radix_str[radix << 1];
   *p = '\0';
}

/* converts a double to string using the current radix
   sets also the edit_radix for possible edit mode
*/
void dtostr( double d, bool_t show_radix ) {

   if( radix < RADIX_16 ) {
      // Standard floating point mode
      dtostr_10(d);
   } else {
      // Integer mode (hex/oct/bin)
      dtostr_radix( d, show_radix );
   }
}

static void angle_mode (void) {

	const nokstr_t   *p;
	struct font      *font;

	if( stack == NULL || stack->depth < (state & STATE_EDIT ? STACK_ROWS - 1 : STACK_ROWS) ) {
		p = (const nokstr_t*)&angle_units[own_get_config(confid_angle_unit) * 6];
		get_font( "small", "plain", &font );
		own_print_string (p, screen_info.width
			- own_get_string_width (p, font), ICON_LINE, font, DM_SET);
	}
}

/* Draw the DEG/RAD/GRAD indicator
*/
void draw_mode (void) {

	freegfx_begin (FALSE);
	angle_mode ();
	freegfx_end (FALSE);
}

/* Draws the stack
*/
void draw_stack (void) {

	unsigned int rows, level;
	int x, y;

   // Update current radix setting for global use now
   radix = own_get_config(confid_radix) - 1;

	freegfx_begin (TRUE);
	rows = (state & STATE_EDIT ? STACK_ROWS - 2 : STACK_ROWS - 1);
	if (stack_bot > stack_ptr) stack_bot = stack_ptr;
	if (stack_bot + rows < stack_ptr) stack_bot = stack_ptr - rows;
	level = stack_bot;

	for (y = rows * ROW_HEIGHT + BASE_HEIGHT; y >= 0; y -= ROW_HEIGHT) {
		x = screen_info.width;
		if( stack != NULL && level < stack->depth ) {
			dtostr( stack->stack[level], true );
			x -= own_get_string_width( (const nokstr_t*)buf, font1 );
			if (x < 0) x = 0;
			own_print_string( (const nokstr_t*)buf, x, y, font1, DM_SET );
		}

		ntoa (++level, buf);
		strcat (buf, ":");
		if( x > own_get_string_width( (const nokstr_t*)buf, font1 ) ) {
			own_print_string( (const nokstr_t*)buf, 1, y, font1, DM_SET );
		}
	}

	angle_mode ();

	if (state & STATE_STACK) {
		// Draw highlight on stack->stack
        own_draw_box (0, (rows - stack_ptr + stack_bot) * ROW_HEIGHT,
			screen_info.width, ROW_HEIGHT, DM_INVERT);
	}
	freegfx_end (FALSE);
}

/* (Re)set the action
*/
void draw_action (void) {

	unsigned long action;

	action = action_options_back;
	if (state & STATE_STACK) {
		if (state & STATE_EDIT) {
			action = (edit_ptr == 0 ? action_ok_back : action_ok_clear);
		}
	} else if (state & STATE_EDIT) {
		if (edit_ptr != 0) action = action_options_clear;
	} else {
		if( stack != NULL ) action = action_options_drop;
	}
	set_action (action);
}

/* Draw cursor (if not waiting for hex letter) and (re)set timer
*/
static unsigned int draw_cursor(void) {

   unsigned int x, t;
   
	x = own_get_string_width( (const nokstr_t*)buf, font1 );
	t = TICKS_PER_SECOND * 2;
	if( wait_hex == 0 ) {
      // Show cursor only if not waiting for hex letter
      own_draw_bitmap( &bmp_cursor, x, EDIT_LINE - BASE_HEIGHT + 1, DM_INVERT );
      x += CURSOR_WIDTH;
      t = TICKS_PER_SECOND / 2;
   }
   own_timer_start( &timer, TT_FUNCTION, t );
   return x;
}

/* Accepts the current digit (if currently entering 2ABC/3DEF.
   Returns true if there were no digit to accept.
*/
bool_t accept(void) {

   if( wait_hex == 0 ) return true;
   // Accept the digit
   wait_hex = 0;
   draw_edit();
   return false;
}

/* Called to blink the cursor or timeout a hexadecimal digit
*/
void timer(void) {

   if( !accept() ) return;
   freegfx_begin(false);
   draw_cursor();
   freegfx_end(false);
}

/* Draws the command line (edit)
*/
void draw_edit(void) {

   unsigned int x;

	draw_action();
	freegfx_begin(false);
	own_clear_box( 0, EDIT_LINE - BASE_HEIGHT + 1, screen_info.width, ROW_HEIGHT );
	
	own_print_string( (const nokstr_t*)buf, 0, EDIT_LINE, font1, DM_SET );
	x = draw_cursor();
   
   if( buf[0] == '#' ) {
      // Append radix letter behind the cursor if integer mode
      own_print_string( (const nokstr_t*)&radix_str[radix << 1],
         x, EDIT_LINE, font1, DM_SET );
   }
	freegfx_end(false);
}

