/*
** edit.c
**
** Part of RPN calculator
** Copyright (c) 2005, 2006, 2007, 2008, 2010 NuukiaWorld
**
**  RPN calculator is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* includes nokix headers */
#include <nokix/all.h>
#include <math.h>
#include "float.h"
#include <ctype.h>

#include "main.h"
#include "stack.h"
#include "draw.h"
#include "calc.h"

/*******************************************************/

/*******************************************************/

/*******************************************************/

static bool_t enter_char (char_t c) {

 	if ((state & STATE_EDIT) == 0) {
  		// Start edit mode
  		state |= STATE_EDIT;
  		edit_ptr = sign_ptr = 0;
  		draw_stack ();
      // In DEC mode, integers default to HEX
      // radix will be restored at draw_stack() when leaving edit mode
      if( radix < RADIX_16 ) radix = RADIX_16;
 	}
 	if( edit_ptr >= ( buf[sign_ptr] == '-' || buf[sign_ptr] == '+'
      ? MAX_EDIT : MAX_EDIT - 1 ) - ( sign_ptr == 0 ? 1 : 0 ) ) {

    	return false;
 	}
 	buf[edit_ptr++] = c;
 	buf[edit_ptr] = '\0';
   draw_edit();
   return true;
}

void change_sign(void) {

 	switch (buf[sign_ptr]) {
      case '-':
  			buf[sign_ptr] = '+';
  			// FALL-THROUGH
    	case '#':
         break;
         
      default:
  			if (buf[sign_ptr] != '+') {
  				memmove (&buf[sign_ptr + 1], &buf[sign_ptr],
               ++edit_ptr - sign_ptr);
    		}
  			buf[sign_ptr] = '-';
 	}
 	draw_edit();
}

void edit_hash(void) {

   if( !( state & STATE_EDIT ) || edit_ptr == 0 ) {
      // Begin integer entry
      enter_char('#');
      return;
   }
   if( buf[0] == '#' ) {
      // Change entry radix
      radix = ( radix + 1 ) % 3;
      draw_edit();
      return;
   }
   if( strchr( buf, '.' ) ) {
		// E
		if( sign_ptr == 0 && enter_char('E') ) sign_ptr = edit_ptr;
		return;
   }
   // .
   enter_char('.');
}

/* Check the pressed key and rotates last hexadecimal digit
   when applicable. Returns true if a digit was rotated.
*/
bool_t edit_rotate(void) {

   char_t c;

   if( wait_hex == 0 ) return false;
   if( wait_hex != message.argv[0] ) return accept();
   
   // Rotate the digit
   c = buf[edit_ptr - 1];
   switch(c) {
      case '2':   c = 'A'; break;
      case 'C':   c = '2'; break;
      case '3':   c = 'D'; break;
      case 'F':   c = '3'; break;
      default:    c++;
   }
   buf[edit_ptr - 1] = c;
   draw_edit();
   return true;
}

void edit_digit( unsigned int digit ) {

   if( buf[0] == '#' ) {
      // Ignore too large digits for current edit_radix
      if( digit > digit_max[radix] ) return;
      
      // Begin rotating 2ABC and 3DEF in hexadecimal mode
      if( radix == RADIX_16 && ( digit == 2 || digit == 3 ) ) {
          wait_hex = digit;
      }
   }
   
   // Cancel rotating if character does not fit in
   if( !enter_char('0' + digit) ) wait_hex = 0;
}

/* Checks the syntax of the entered value
*/
static bool_t check_syntax(void) {

	unsigned int c;

   if( buf[0] == '#' ) {
      // Integer value
      if( buf[1] == '\0' ) return false;
      for (c = 1; c < edit_ptr; c++) {
         if( buf[c] > digit_max_char[radix] ) return false;
      }
      return true;
   }

   // Check the exponent or mantissa
   // must be at least one numeric character
   for (c = sign_ptr; c < edit_ptr; c++) {
      if (isdigit (buf[c])) {
      	// Syntax ok if no exponent
        	if (sign_ptr == 0) return true;
         // Exponent ok, check the mantissa
         // must be at least one numeric character
         for (c = 0; c < sign_ptr; c++) {
         	if (isdigit (buf[c])) return true;
         }
         return false;
      }
   }
   return false;
}

/* Converts floating point string to double
*/
static double strtod_10(void) {

   double d;
	int exp;

   // Simple atof() call does not work because it does not handle full double range
	if (sign_ptr != 0) buf[sign_ptr - 1] = '\0';
	d = atof(buf);   // Fetches mantissa
	if (sign_ptr != 0) {
		buf[sign_ptr - 1] = 'E';
		if( d == 0 ) return d;   // Zero mantissa, ignore exponent
      // NON-STANDARD: double -> int conversion overflow expected to clamp
      exp = (int)atof( &buf[sign_ptr] ); // atol() range not enough!
  		if( exp < 0 ) {
/*			while (exp <= -16 && d != 0) {
				d /= 1E16;
				exp += 16;
			}*/
			while (exp++ < 0 && d != 0) d /= 10;
  		} else {
/*   		while (exp >= 16 && fabs (d) <= DBL_MAX) {
   			d *= 1E16;
   			exp -= 16;
   		}*/
   		while (exp-- > 0 && fabs (d) <= DBL_MAX) d *= 10;
  		}
	}
	return d;
}

/* Converts integer string in edit_radix to double
*/
static double strtod_radix(void) {

	unsigned int digit, l;
   char *p;

   l = 0;
   for (p = &buf[1]; *p != '\0'; p++) {
   	digit = *p - '0';
      if (digit > 9) digit -= 'A' - '9' - 1;
   	l = ( l << digit_bits[radix] ) + digit;
   }
   return (double)(signed int)l;
}

/* Pushes edited value to the stack->stack, returns TRUE and cancels
   edit state if successful
*/
bool_t edit_enter(void) {

	double d;

   if( !(state & STATE_EDIT ) ) return true;
   if( edit_ptr != 0 ) {
      // Check entered value
    	if( !check_syntax() ) {
         display_msg( DMI_ERROR, &str_syntax );
        	return false;
      }
    	// Decode entered value
      if( buf[0] == '#' ) {
        	d = strtod_radix();     // Integer
      } else {
         d = strtod_10();        // Floating point
         if( !check_value(d) ) return false;
      }
    	// Add entered value to the stack
      if( !stack_dup( ( state & STATE_STACK ) ? 0 : 1 ) ) return false;
    	stack->stack[stack_ptr] = d;
   }
   own_timer_abort(&timer);
	state &= ~STATE_EDIT;
	draw_action();
	return true;
}

