/*
** main.c
**
** Part of RPN calculator
** Copyright (c) 2005, 2006, 2007, 2008, 2009, 2010 NuukiaWorld
**
**  RPN calculator is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* includes nokix headers */
#include <nokix/all.h>
extern	void	menu_reset (void);

#include "main.h"
#include "edit.h"
#include "stack.h"
#include "draw.h"
#include "calc.h"
#include "conv.h"
#include "const.h"

/*******************************************************/

/* Calculation types for OMSG_CALC
   0x0000...0x00FF   2-argument operators
   0x0100...0x01FF   1-argument functions
   0x0200...0x02FF   0-argument constants
   0x0300...0x03FF   N-argument statistics
*/
static void(*calcs[])(void) = { calc_operator, calc_function, calc_constant, calc_statistic };

/*******************************************************/

/*******************************************************/

/* Called whenever the options menu is closed on other than
   root level
*/
static void menu_exit(void) {
   menu_reset();                    // Remove menus from "menu stack->stack"
	state &= ~( STATE_MENU | STATE_CONST );
   draw_action();                   // Activate the correct action
}

static void stack_exit(void) {
	state &= ~STATE_STACK;
	stack_ptr = 0;
   draw_action();
}

/* messages dispatcher */
unsigned int dispatcher(unsigned int msg) {

 switch(msg) {

	case MSG_D_INIT:
		freegfx_open_ex( FGF_SHOW_ACTION );
		menu_create (1, MDM_SUB | MDM_LIST, MCM_MARK_ID, 1, NULL);
		srand(get_timestamp());
		const_mode = wait_hex = state = stack_ptr = 0;
		last_stack = stack;            // Undo not available
		get_font ("small", "bold", &font1);
		draw_stack ();
		msg = MSG_R_STAY;
      // FALL-THROUGH
   case MSG_CALL_FINISHED:
		draw_action (); // Restore application's actions after a call
		break;

	case MSG_D_QUIT:
		if (msg_argv[0] == 3 || !(state & STATE_MENU)) {
        	// Exit application
        	own_timer_abort( &timer );
        	if( last_stack != stack ) free(last_stack);  // Keep stack, discard undo
			menu_delete (1);
			freegfx_close ();
			msg = MSG_R_QUIT;
			break;
		}
		// Closing the menu because of
        // - Invalid shortcut entered
        // - Angle unit selected with shortcut
        // - Back pressed in root level
		menu_close ();
      menu_exit();
		if( state & STATE_EDIT ) {
         draw_mode();					// Update DEG/RAD/GRAD indicator
      } else {
         draw_stack();              // Radix may have been changed
      }
      break;

	case MSG_D_QUIT_NOW:
		// Acts the same way as the built-in calculator:
		msg = MSG_R_STAY;
		switch (msg_argv[0]) {
			case 1:							// Exchange rate set
            menu_exit();
				break;
			case 2:    						// Exchange rate canceled
				menu_reopen();
				break;
			case 3:							// Exit app., no idea when happens
				msg = MSG_R_LEAVE;
				break;
		}
		break;

	case MSG_BACK:
		state &= ~( STATE_SHORTCUT | STATE_CONST );   // Shortcut session closed
		conv_unit = -1;						// Clear any selected conversion
      conv_flags (0);						// Restore coversion menu flags
		if (state == 0 || (state & STATE_MENU)) break;
		if (state & STATE_EDIT) {			// Cancel edit mode
         edit_ptr = 0;
        	edit_enter();
      } else {                         // Exit stack mode
			stack_exit();
		}
      draw_stack ();
		msg = MSG_NOP;
		break;

	case MSG_KEY_UP:
		state &= ~STATE_SHORTCUT;			// Shortcut session closed
		if (state & STATE_MENU) break;
		if (state & STATE_EDIT) {
         if( !accept() ) break;
         if( !edit_enter() ) break;
      }
		if (state & STATE_STACK) {
         stack_ptr++;
			if( stack_ptr >= stack->depth ) stack_ptr = 0;
		} else {
			if( stack == NULL ) {
				display_msg( DMI_INFORMATION, &str_empty );
			} else {
				// Start stack->stack mode
				state |= STATE_STACK;
				draw_action ();
			}
		}
		draw_stack ();
		break;

  	case MSG_KEY_DOWN:
		state &= ~STATE_SHORTCUT;			// Shortcut session closed
		if (state & STATE_MENU) break;
		if (state & STATE_EDIT) {
         if( !accept() ) break;
			if( !edit_enter() ) break;
			if (state & STATE_STACK) {
			   if( stack_ptr == 0 ) stack_ptr = stack->depth;
			   stack_ptr--;
			}
		} else if (state & STATE_STACK) {
			if( stack_ptr == 0 ) stack_ptr = stack->depth;
			stack_ptr--;
		} else {
			// DUP
			if (!stack_arg_check (1)) break;
			stack_dup (1);
		}
		draw_stack ();
		break;

	case MSG_SELECT:
		state &= ~STATE_SHORTCUT;			// Shortcut session closed
		if (state & STATE_MENU) break;
		if (edit_enter ()) draw_stack ();// OK pressed while editing stack
		break;

	case MSG_KEYPRESS:
      if( ( state & STATE_CONST ) && msg_argv[0] == KEY_HASH ) const_switch();
		if( state & STATE_MENU ) break;
		
      // MSG_KEY_UP and MSG_KEY_DOWN processes the arrow keys
      if( msg_argv[0] == KEY_UP || msg_argv[0] == KEY_DOWN ) break;

      if( edit_rotate() ) break;
		if( msg_argv[0] < KEY_1 ) break;
		
		if( ( state & STATE_EDIT ) || !( state && STATE_STACK ) ) {
			switch (msg_argv[0]) {
				case KEY_ASTERISK:
               if( state & STATE_EDIT ) {
                  // +/-
                  change_sign();
					} else if( stack_arg_check(1) ) {
			         // NEG
			         stack->stack[0] = -stack->stack[0];
                  draw_stack();
               }
               break;
					
				case KEY_HASH:
               // # . E
               edit_hash();
					break;
					
				case KEY_0:
               // 0
					edit_digit(0);
					break;
					
				default:
					// 1...9
					if( msg_argv[0] <= KEY_9 ) edit_digit(msg_argv[0]);
			}
			break;
		}
      if( ( state & STATE_STACK ) && msg_argv[0] <= KEY_9
         && stack_arg_check( msg_argv[0] - KEY_1 + 1 ) ) {

			// Roll shortcut (n = 1...9), exits stack->stack mode
			stack_roll (msg_argv[0] - KEY_1);
			stack_exit();
         draw_stack ();
      }
		break;

	case MSG_CLEAR:
		if( edit_ptr != 0 ) {
			if (sign_ptr == edit_ptr) sign_ptr = 0;	// deleting E
			buf[--edit_ptr] = '\0';
			draw_edit ();
		}
		msg = MSG_NOP;
		break;                     

	case MSG_KEYPRESS_ASCII:
		// Ignore number keys (also in menu like in the built-in calculator)
		// However, allow them if menu shortcut session is open
		if (!(state & STATE_SHORTCUT)) msg = MSG_NOP;
		break;

	case MSG_DIALOG_ACCEPTED:
		// Detects the shortcut timeout, this seems to work...
		// No idea about the actual meaning of argv[0] values here
		if (msg_argv[0] == 1) state &= ~STATE_SHORTCUT;
		break;

	case OMSG_CALC:
		if( edit_enter() ) {
         calcs[msg_argv[0] >> 8 ]();
			draw_stack ();
		}
      menu_exit();
		msg = MSG_NOP;
		break;

	case OMSG_STACK:
		stack_operation ();
      menu_exit();
		msg = MSG_NOP;
		break;

	case MSG_CLEAR_ALL:
		state &= ~STATE_MENU;	// May be selected also in the main menu
		if (state & STATE_EDIT) {
			// Clear the command line only
			edit_ptr = sign_ptr = 0;
			buf[0] = '\0';
			draw_edit ();
		} else {
			// CLEAR
			stack_set_depth(0);
			draw_stack ();
		}
		msg = MSG_NOP;
		break;

   case OMSG_UNDO:
		if( last_stack != stack ) {    // Only if undo available
   		if( state & STATE_EDIT ) {  // Cancel edit
            edit_ptr = 0;
           	edit_enter();
         }
         stack_undo();
         stack_ptr_check();
         draw_stack();
      }
      menu_exit();
      msg = MSG_NOP;
		break;

	case MSG_OPTIONS:
      menu_exit();            // Delete previous menu state
		state |= STATE_MENU;
		send_message( MSG_OPEN_STATIC_MENU|MSG_1P,
			( state & STATE_STACK ? &menu_stack : &menu_options ) );
		msg = MSG_NOP;
		break;

	case OMSG_DROP:
		stack_drop ();
		draw_stack ();
		msg = MSG_NOP;
		break;

	case 0x10CE:
	case 0x10CF:
		// Set exchange rate
		// Edit mode must be closed because same buffer is used
		if( state & STATE_EDIT ) {
         if( !edit_enter() ) {
            menu_exit();
            msg = MSG_NOP;
            break;
         }
         draw_stack();
      }
      send_message (MSG_START_HANDLER|MSG_2P, id_exchange_rate, msg & 1);
		msg = MSG_NOP;
		break;

	case OMSG_EXCHANGE:
		// Shortcut session is ended when entering exchange rate submenu
		state &= ~STATE_SHORTCUT;
		send_message (MSG_OPEN_STATIC_MENU|MSG_1P, msg_argv[0]);
		msg = MSG_NOP;
		break;

	case OMSG_CONST_MENU:
		// Constants group selected
		state |= STATE_CONST;
      const_init();
		menu_open( 1, 1 );
		msg = MSG_NOP;
		break;

	case OMSG_CONST:
      if( edit_enter() ) {
         const_calc();
         draw_stack();
      }
      menu_close();
      menu_exit();
		msg = MSG_NOP;
		break;

	case OMSG_CONV_MENU:
		// Unit conversion group selected
		conv_init ();
		menu_open (1, 1);
		msg = MSG_NOP;
		break;

	case OMSG_CONV_UNIT:
		if (conv_unit < 0) {
			// "Convert from" unit selected, open the "Convert to" menu
			conv_unit = msg_argv[0];
         display_msg( DMI_DONE, &str_select_dest_unit );
         conv_flags (MIF_LEAVE_MENU);	// same menu but different flags
			menu_open (1, 1);
		} else {
			// "Convert to" unit selected, perform the conversion
			if( edit_enter() ) {
				conv_calc ();
				draw_stack ();
			}
         menu_close ();
         menu_exit();
		}
		msg = MSG_NOP;
		break;
 }
 return msg;
}

